#!/usr/bin/env python

"""
SVG-specific document support.
See: http://www.w3.org/TR/SVGMobile12/python-binding.html
See: http://www.w3.org/TR/SVGMobile12/svgudom.html

Copyright (C) 2007 Paul Boddie <paul@boddie.org.uk>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

import libxml2dom
from libxml2dom.events import *
from libxml2dom.macrolib import *
import xml.dom
import math

SVG_NAMESPACE = "http://www.w3.org/2000/svg"

class _Exception(Exception):

    "A generic SVG exception."

    def __init__(self, code):
        Exception.__init__(self, code)
        self.code = code

class SVGException(_Exception):

    "An SVG exception."

    SVG_WRONG_TYPE_ERR = 0
    SVG_INVALID_VALUE_ERR = 1
    SVG_MATRIX_NOT_INVERTABLE = 2

class GlobalException(_Exception):

    "A global exception."

    NOT_CONNECTED_ERR = 1
    ENCODING_ERR = 2
    DENIED_ERR = 3
    UNKNOWN_ERR = 4

class SVGImplementation(libxml2dom.Implementation):

    "Contains an SVG-specific implementation."

    # Wrapping of documents.

    def adoptDocument(self, node):
        return SVGDocument(node, self)

    # Factory functions.

    def get_node(self, _node, context_node):
        if Node_nodeType(_node) == context_node.ELEMENT_NODE and \
            Node_namespaceURI(_node) == SVG_NAMESPACE:

            if Node_localName(_node) == "svg":
                return SVGSVGElement(_node, self, context_node.ownerDocument)
            else:
                return SVGElement(_node, self, context_node.ownerDocument)
        else:
            return libxml2dom.Implementation.get_node(self, _node, context_node)

# Interfaces and helper classes.

class AsyncStatusCallback:

    "An asynchronous callback interface."

    def operationComplete(self, status):
        pass

class AsyncURLStatus:

    "The status of a URL retrieval operation."

    def __init__(self, success, contentType, content):
        self.success, self.contentType, self.content = success, contentType, content

class SVGGlobal: # Global, EventListenerInitializer2

    "An SVG global."

    def __init__(self, document): # parent

        "Initialise the global with the given 'document'."

        self.document = document

    def createConnection(self):
        raise NotImplementedError, "createConnection"

    def createTimer(self, initialInterval, repeatInterval):
        raise NotImplementedError, "createTimer"

    def gotoLocation(self, newIRI):
        raise NotImplementedError, "gotoLocation"

    def binaryToString(self, octets, encoding):
        raise NotImplementedError, "binaryToString"

    def stringToBinary(self, data, encoding):
        raise NotImplementedError, "stringToBinary"

    def getURL(self, iri, callback):

        # NOTE: Not asynchronous.
        # NOTE: The urlopen function may not support IRIs.
        # No exceptions are supposed to be raised, which is a bit nasty.

        f = urllib.urlopen(iri)
        try:
            try:
                content = f.read()
                contentType = f.headers["Content-Type"]
                callback.operationComplete(AsyncURLStatus(1, contentType, content))
            except:
                callback.operationComplete(AsyncURLStatus(0, None, None))
        finally:
            f.close()

    def postURL(self, iri, data, callback, type, encoding):
        raise NotImplementedError, "postURL"

    def parseXML(self, data, contextDoc):
        doc = parseString(data)
        return contextDoc.importNode(doc.documentElement, 1)

class SVGLocatable:

    "A locatable interface."

    pass

class SVGMatrix:

    """
    A matrix.
    See: http://www.w3.org/TR/SVGMobile12/svgudom.html#svg__SVGMatrix
    """

    def __init__(self, a, b, c, d, e, f):
        self.matrix = a, b, c, d, e, f

    def getComponent(self, index):

        """
        Return the component with the given 'index' (starting at zero) from the
        sequence a, b, c, d, e, f where each element corresponds to the matrix
        as follows:

        [ a c e ]
        [ b d f ]
        [ 0 0 1 ]
        """

        try:
            return self.matrix[index]
        except IndexError:
            raise xml.dom.DOMException(xml.dom.INDEX_SIZE_ERR)

    def mMultiply(self, secondMatrix):

        """
        Multiply this matrix with 'secondMatrix' and update its contents to the
        result of the multiplication operation defined as follows:

        [ a c e ] [ A C E ]
        [ b d f ] [ B D F ]
        [ 0 0 1 ] [ 0 0 1 ]

        Return this object as a result.
        """

        a, b, c, d, e, f = self.matrix
        A, B, C, D, E, F = secondMatrix.matrix
        self.matrix = a*A + c*B, b*A + d*B, a*C + c*D, b*C + d*D, a*E + c*F + e, b*E + d*F + f
        return self

    def inverse(self):

        """
        det = ad - cb

        See (for example): http://mathworld.wolfram.com/MatrixInverse.html
        """

        det = a*d - c*b
        if det != 0:
            m = 1/det
            a, b, c, d, e, f = self.matrix
            self.matrix = m * d, m * -b, m * -c, m * a, m * (c*f - e*d), m * (e*b - a*f)
            return self
        else:
            raise SVGException(SVGException.SVG_MATRIX_NOT_INVERTABLE)

    def mTranslate(self, x, y):

        """
        [ 1 0 x ]
        [ 0 1 y ]
        [ 0 0 1 ]
        """

        return self.mMultiply(SVGMatrix(1, 0, 0, 1, x, y))

    def mScale(self, scaleFactor):

        """
        [ scaleFactor 0           0 ]
        [ 0           scaleFactor 0 ]
        [ 0           0           1 ]
        """

        return self.mMultiply(SVGMatrix(scaleFactor, 0, 0, scaleFactor, 0, 0))

    def mRotate(self, angle):

        """
        [ cos(angle) -sin(angle) 0 ]
        [ sin(angle) cos(angle)  0 ]
        [ 0          0           1 ]
        """

        return self.mMultiply(SVGMatrix(math.cos(angle), math.sin(angle), -math.sin(angle), math.cos(angle), 0, 0))

class SVGPath:

    """
    A path.
    See: http://www.w3.org/TR/SVGMobile12/svgudom.html#svg__SVGPath
    See: http://www.w3.org/TR/SVGMobile12/paths.html
    """

    MOVE_TO = 77
    LINE_TO = 76
    CURVE_TO = 67
    QUAD_TO = 81
    CLOSE = 90

    def __init__(self):
        self.segments = []

    def _numberOfSegments(self):
        return len(self.segments)

    numberOfSegments = property(_numberOfSegments)

    def getSegment(self, cmdIndex):
        try:
            return self.segments[cmdIndex][0]
        except IndexError:
            raise xml.dom.DOMException(xml.dom.INDEX_SIZE_ERR)

    def getSegmentParam(self, cmdIndex, paramIndex):
        try:
            return self.segments[cmdIndex][1][paramIndex]
        except IndexError:
            raise xml.dom.DOMException(xml.dom.INDEX_SIZE_ERR)

    def moveTo(self, x, y):
        self.segments.append((self.MOVE_TO, (x, y)))

    def lineTo(self, x, y):
        self.segments.append((self.LINE_TO, (x, y)))

    def quadTo(self, x1, y1, x2, y2):
        self.segments.append((self.QUAD_TO, (x1, y1, x2, y2)))

    def curveTo(self, x1, y1, x2, y2, x3, y3):
        self.segments.append((self.CURVE_TO, (x1, y1, x2, y2, x3, y3)))

    def close(self):
        self.segments.append((self.CLOSE,))

class SVGPoint:

    "A point."

    def __init__(self, x, y):
        self.x = x
        self.y = y

class SVGRect:

    "A rectangle."

    def __init__(self, x, y, width, height):
        self.x, self.y, self.width, self.height = x, y, width, height

class SVGRGBColor:

    "A colour."

    def __init__(self, red, green, blue):
        self.red, self.green, self.blue = red, green, blue

# Node classes.

class SVGDocument(libxml2dom.Document, DocumentEvent, EventTarget):

    "An SVG-specific document node."

    def __init__(self, node, impl):

        """
        Initialise the document with the given 'node', implementation 'impl',
        and global (SVGGlobal) details.
        """

        libxml2dom.Document.__init__(self, node, impl)
        self.global_ = SVGGlobal(self) # parent

class SVGElement(libxml2dom.Node, EventTarget): # (Element), TraitAccess, ElementTraversal

    "An SVG-specific element."

    def _id(self):
        return self.getAttribute("id")

    def _setId(self, value):
        self.setAttribute("id", value)

    id = property(_id, _setId)

class SVGLocatableElement(SVGElement, SVGLocatable):

    "A locatable element."

    pass

class SVGTimedElement(SVGElement): # smil::ElementTimeControl

    "A timed element."

    def __init__(self, *args):

        "Initialise the element with the underlying 'args'."

        SVGElement.__init__(self, *args)
        self.document_time = 0
        self.paused = 0

    def _isPaused(self):
        return self.paused

    def pauseElement(self):
        self.paused = 1

    def resumeElement(self):
        self.paused = 0

class SVGSVGElement(SVGLocatableElement, SVGTimedElement):

    "An SVG-specific top-level element."

    NAV_AUTO = 1
    NAV_NEXT = 2
    NAV_PREV = 3
    NAV_UP = 4
    NAV_UP_RIGHT = 5
    NAV_RIGHT = 6
    NAV_DOWN_RIGHT = 7
    NAV_DOWN = 8
    NAV_DOWN_LEFT = 9
    NAV_LEFT = 10
    NAV_UP_LEFT = 11

    def __init__(self, *args):

        "Initialise the element with the underlying 'args'."

        SVGTimedElement.__init__(self, *args)
        self.scale = 1
        self.rotate = 0
        self.translate = SVGPoint(0, 0)

    def _currentScale(self):
        return self.scale

    def _currentRotate(self):
        return self.rotate

    def _currentTranslate(self):
        return self.translate

    def _viewport(self):
        attr = self.getAttribute("viewBox")
        if attr is None:
            return attr
        l = map(int, attr.split())
        return SVGRect(*l)

    def getCurrentTime(self):
        return self.document_time

    def setCurrentTime(self, setCurrentTime):
        self.document_time = setCurrentTime

    def createSVGMatrixComponents(self, a, b, c, d, e, f):
        return SVGMatrix(a, b, c, d, e, f)

    def createSVGRect(self):
        return SVGRect(0, 0, 0, 0)

    def createSVGPath(self):
        return SVGPath()

    def createSVGRGBColor(self, red, green, blue):
        return SVGRGBColor(red, green, blue)

    def moveFocus(self, motionType):
        raise NotImplementedError, "moveFocus"

    def setFocus(self, object):
        raise NotImplementedError, "setFocus"

    def getCurrentFocusedObject(self):
        raise NotImplementedError, "getCurrentFocusedObject"

    currentScale = property(_currentScale)
    currentRotate = property(_currentRotate)
    currentTranslate = property(_currentTranslate)
    viewport = property(_viewport)

# Convenience functions.

def parse(stream_or_string, html=0, htmlencoding=None):
    return libxml2dom.parse(stream_or_string, html, htmlencoding, default_impl)

def parseFile(filename, html=0, htmlencoding=None):
    return libxml2dom.parseFile(filename, html, htmlencoding, default_impl)

def parseString(s, html=0, htmlencoding=None):
    return libxml2dom.parseString(s, html, htmlencoding, default_impl)

def parseURI(uri, html=0, htmlencoding=None):
    return libxml2dom.parseURI(uri, html, htmlencoding, default_impl)

# Single instance of the implementation.

default_impl = SVGImplementation()

# vim: tabstop=4 expandtab shiftwidth=4
